use serde::Serialize;
use super::value::writer::Writer;
use super::value::{
    Value, NestedValue, PlainValue,
    Domain, NullDomain,
    EncodePlaceholderMap,
    Encoder,
};

pub use super::error::Error;
type Result<T> = std::result::Result<T, Error>;

#[derive(Debug)]
pub struct Serializer<'a, W: Writer, N: NestedValue<D>, D: Domain> {
    pub write: &'a mut W,
    placeholders: Option<&'a EncodePlaceholderMap<N, D>>,
}

impl<'a, W: Writer, N: NestedValue<D>, D: Domain> Serializer<'a, W, N, D> {
    pub fn new(write: &'a mut W, placeholders: Option<&'a EncodePlaceholderMap<N, D>>) -> Self {
        Serializer { write, placeholders }
    }

    fn write_symbol(&mut self, s: &str) -> Result<()> {
        match self.placeholders.as_ref().and_then(|m| m.get(&Value::symbol(s))) {
            Some(&n) => Ok(self.write.write_placeholder_ref(n)?),
            None => Ok(self.write.write_symbol(s)?),
        }
    }
}

#[derive(Debug)]
pub struct SerializeCompound<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> {
    ser: &'a mut Serializer<'b, W, N, D>,
    count: Option<usize>,
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::Serializer for &'a mut Serializer<'b, W, N, D> {
    type Ok = ();
    type Error = Error;
    type SerializeSeq = SerializeCompound<'a, 'b, W, N, D>;
    type SerializeTuple = SerializeCompound<'a, 'b, W, N, D>;
    type SerializeTupleStruct = SerializeCompound<'a, 'b, W, N, D>;
    type SerializeTupleVariant = SerializeCompound<'a, 'b, W, N, D>;
    type SerializeMap = SerializeCompound<'a, 'b, W, N, D>;
    type SerializeStruct = SerializeCompound<'a, 'b, W, N, D>;
    type SerializeStructVariant = SerializeCompound<'a, 'b, W, N, D>;

    fn serialize_bool(self, v: bool) -> Result<()> {
        Ok(self.write.write_bool(v)?)
    }

    fn serialize_i8(self, v: i8) -> Result<()> {
        Ok(self.write.write_i8(v)?)
    }

    fn serialize_i16(self, v: i16) -> Result<()> {
        Ok(self.write.write_i16(v)?)
    }

    fn serialize_i32(self, v: i32) -> Result<()> {
        Ok(self.write.write_i32(v)?)
    }

    fn serialize_i64(self, v: i64) -> Result<()> {
        Ok(self.write.write_i64(v)?)
    }

    fn serialize_u8(self, v: u8) -> Result<()> {
        Ok(self.write.write_u8(v)?)
    }

    fn serialize_u16(self, v: u16) -> Result<()> {
        Ok(self.write.write_u16(v)?)
    }

    fn serialize_u32(self, v: u32) -> Result<()> {
        Ok(self.write.write_u32(v)?)
    }

    fn serialize_u64(self, v: u64) -> Result<()> {
        Ok(self.write.write_u64(v)?)
    }

    fn serialize_f32(self, v: f32) -> Result<()> {
        Ok(self.write.write_f32(v)?)
    }

    fn serialize_f64(self, v: f64) -> Result<()> {
        Ok(self.write.write_f64(v)?)
    }

    fn serialize_char(self, v: char) -> Result<()> {
        self.write.open_record(1)?;
        self.write_symbol("UnicodeScalar")?;
        self.write.write_u32(v as u32)?;
        Ok(self.write.close_record()?)
    }

    fn serialize_str(self, v: &str) -> Result<()> {
        Ok(self.write.write_string(v)?)
    }

    fn serialize_bytes(self, v: &[u8]) -> Result<()> {
        Ok(self.write.write_bytes(v)?)
    }

    fn serialize_none(self) -> Result<()> {
        self.write.open_record(0)?;
        self.write_symbol("None")?;
        Ok(self.write.close_record()?)
    }

    fn serialize_some<T: ?Sized>(self, v: &T) -> Result<()> where T: Serialize {
        self.write.open_record(1)?;
        self.write_symbol("Some")?;
        v.serialize(&mut *self)?;
        Ok(self.write.close_record()?)
    }

    fn serialize_unit(self) -> Result<()> {
        self.write.open_record(0)?;
        self.write_symbol("tuple")?;
        Ok(self.write.close_record()?)
    }

    fn serialize_unit_struct(self, name: &'static str) -> Result<()> {
        self.write.open_record(0)?;
        self.write_symbol(name)?;
        Ok(self.write.close_record()?)
    }

    fn serialize_unit_variant(self,
                              _name: &'static str,
                              _variant: u32,
                              variant_name: &'static str) ->
        Result<()>
    {
        self.write.open_record(0)?;
        self.write_symbol(variant_name)?;
        Ok(self.write.close_record()?)
    }

    fn serialize_newtype_struct<T: ?Sized>(self, name: &'static str, value: &T) ->
        Result<()> where T: Serialize
    {
        if name == crate::value::value::MAGIC_VALUE {
            let b = unsafe { Box::from_raw(*((value as *const T) as *const u64) as *mut &Value<N, D>) };
            let v: &Value<N, D> = *b;
            Encoder::new(self.write, self.placeholders).write_value(v)?;
            Ok(())
        } else if name == crate::value::value::MAGIC_NESTEDVALUE {
            let b = unsafe { Box::from_raw(*((value as *const T) as *const u64) as *mut &N) };
            let v: &N = *b;
            Encoder::new(self.write, self.placeholders).write(v)?;
            Ok(())
        } else {
            // TODO: This is apparently discouraged, and we should apparently just serialize `value`?
            self.write.open_record(1)?;
            self.write_symbol(name)?;
            value.serialize(&mut *self)?;
            Ok(self.write.close_record()?)
        }
    }

    fn serialize_newtype_variant<T: ?Sized>(self,
                                            _name: &'static str,
                                            _variant: u32,
                                            variant_name: &'static str,
                                            value: &T) ->
        Result<()> where T: Serialize
    {
        self.write.open_record(1)?;
        self.write_symbol(variant_name)?;
        value.serialize(&mut *self)?;
        Ok(self.write.close_record()?)
    }

    fn serialize_seq(self, count: Option<usize>) -> Result<Self::SerializeSeq> {
        match count {
            Some(n) => self.write.open_sequence(n)?,
            None => self.write.stream_sequence()?,
        }
        Ok(SerializeCompound { ser: self, count })
    }

    fn serialize_tuple(self, count: usize) -> Result<Self::SerializeTuple> {
        self.write.open_record(count)?;
        self.write_symbol("tuple")?;
        Ok(SerializeCompound { ser: self, count: Some(count) })
    }

    fn serialize_tuple_struct(self, name: &'static str, count: usize) ->
        Result<Self::SerializeTupleStruct>
    {
        self.write.open_record(count)?;
        self.write_symbol(name)?;
        Ok(SerializeCompound { ser: self, count: Some(count) })
    }

    fn serialize_tuple_variant(self,
                               _name: &'static str,
                               _variant: u32,
                               variant_name: &'static str,
                               count: usize) ->
        Result<Self::SerializeTupleVariant>
    {
        self.write.open_record(count)?;
        self.write_symbol(variant_name)?;
        Ok(SerializeCompound { ser: self, count: Some(count) })
    }

    fn serialize_map(self, count: Option<usize>) -> Result<Self::SerializeMap> {
        match count {
            Some(n) => self.write.open_dictionary(n)?,
            None => self.write.stream_dictionary()?,
        }
        Ok(SerializeCompound { ser: self, count })
    }

    fn serialize_struct(self, name: &'static str, count: usize) -> Result<Self::SerializeStruct> {
        self.write.open_record(count)?;
        self.write_symbol(name)?;
        Ok(SerializeCompound { ser: self, count: Some(count) })
    }

    fn serialize_struct_variant(self,
                                _name: &'static str,
                                _variant: u32,
                                variant_name: &'static str,
                                count: usize) ->
        Result<Self::SerializeStructVariant>
    {
        self.write.open_record(count)?;
        self.write_symbol(variant_name)?;
        Ok(SerializeCompound { ser: self, count: Some(count) })
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeMap for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_key<T: ?Sized>(&mut self, key: &T) -> Result<()> where T: Serialize {
        Ok(key.serialize(&mut *self.ser)?)
    }

    fn serialize_value<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_dictionary()?),
        }
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeStruct for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_record()?),
        }
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeStructVariant for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_record()?),
        }
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeTuple for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_record()?),
        }
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeTupleStruct for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_record()?),
        }
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeTupleVariant for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_record()?),
        }
    }
}

impl<'a, 'b, W: Writer, N: NestedValue<D>, D: Domain> serde::ser::SerializeSeq for SerializeCompound<'a, 'b, W, N, D> {
    type Ok = ();
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        Ok(value.serialize(&mut *self.ser)?)
    }

    fn end(self) -> Result<()> {
        match self.count {
            None => Ok(self.ser.write.close_stream()?),
            Some(_) => Ok(self.ser.write.close_sequence()?),
        }
    }
}

pub fn to_writer<W: Writer, T: Serialize>(write: &mut W, value: &T) -> Result<()> {
    let mut ser: Serializer<'_, W, PlainValue<NullDomain>, NullDomain> = Serializer::new(write, None);
    value.serialize(&mut ser)
}
