use crate::value::{Value, NestedValue, Map, Domain};
use serde::Serialize;
use std::marker::PhantomData;

#[derive(Debug)]
pub enum Error {}
impl serde::ser::Error for Error {
    fn custom<T>(_: T) -> Self where T: std::fmt::Display {
        unreachable!()
    }
}
impl std::error::Error for Error {}
impl std::fmt::Display for Error {
    fn fmt(&self, _fmt: &mut std::fmt::Formatter) -> std::result::Result<(), std::fmt::Error> {
        unreachable!()
    }
}

type Result<T> = std::result::Result<T, Error>;

pub struct Serializer<N: NestedValue<D>, D: Domain> {
    phantom: PhantomData<(N, D)>,
}

pub struct SerializeDictionary<N: NestedValue<D>, D: Domain> {
    next_key: Option<N>,
    items: Map<N,N>,
    phantom: PhantomData<D>,
}

pub struct SerializeRecord<N: NestedValue<D>, D: Domain> {
    name: &'static str,
    vec: Vec<N>,
    phantom: PhantomData<D>,
}

pub struct SerializeSequence<N: NestedValue<D>, D: Domain> {
    vec: Vec<N>,
    phantom: PhantomData<D>,
}

impl<N: NestedValue<D>, D: Domain> serde::Serializer for Serializer<N, D> {
    type Ok = N;
    type Error = Error;
    type SerializeSeq = SerializeSequence<N, D>;
    type SerializeTuple = SerializeRecord<N, D>;
    type SerializeTupleStruct = SerializeRecord<N, D>;
    type SerializeTupleVariant = SerializeRecord<N, D>;
    type SerializeMap = SerializeDictionary<N, D>;
    type SerializeStruct = SerializeRecord<N, D>;
    type SerializeStructVariant = SerializeRecord<N, D>;

    fn serialize_bool(self, v: bool) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i8(self, v: i8) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i16(self, v: i16) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i32(self, v: i32) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_i64(self, v: i64) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u8(self, v: u8) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u16(self, v: u16) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u32(self, v: u32) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_u64(self, v: u64) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_f32(self, v: f32) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_f64(self, v: f64) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_char(self, v: char) -> Result<Self::Ok> {
        Ok(Value::simple_record("UnicodeScalar", vec![Value::from(v as u32).wrap()]).wrap())
    }

    fn serialize_str(self, v: &str) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_bytes(self, v: &[u8]) -> Result<Self::Ok> {
        Ok(Value::from(v).wrap())
    }

    fn serialize_none(self) -> Result<Self::Ok> {
        Ok(Value::simple_record("None", vec![]).wrap())
    }

    fn serialize_some<T: ?Sized>(self, v: &T) -> Result<Self::Ok> where T: Serialize {
        Ok(Value::simple_record("Some", vec![to_value(v)]).wrap())
    }

    fn serialize_unit(self) -> Result<Self::Ok> {
        Ok(Value::simple_record("tuple", vec![]).wrap())
    }

    fn serialize_unit_struct(self, name: &'static str) -> Result<Self::Ok> {
        Ok(Value::simple_record(name, vec![]).wrap())
    }

    fn serialize_unit_variant(self,
                              _name: &'static str,
                              _variant: u32,
                              variant_name: &'static str) ->
        Result<Self::Ok>
    {
        Ok(Value::simple_record(variant_name, vec![]).wrap())
    }

    fn serialize_newtype_struct<T: ?Sized>(self, name: &'static str, value: &T) ->
        Result<Self::Ok> where T: Serialize
    {
        if name == crate::value::value::MAGIC_VALUE {
            let b = unsafe { Box::from_raw(*((value as *const T) as *const u64) as *mut &Value<N, D>) };
            let v: &Value<N, D> = *b;
            Ok(v.clone().wrap())
        } else if name == crate::value::value::MAGIC_NESTEDVALUE {
            let b = unsafe { Box::from_raw(*((value as *const T) as *const u64) as *mut &N) };
            let v: &N = *b;
            Ok(v.clone())
        } else {
            // TODO: This is apparently discouraged, and we should apparently just serialize `value`?
            Ok(Value::simple_record(name, vec![to_value(value)]).wrap())
        }
    }

    fn serialize_newtype_variant<T: ?Sized>(self,
                                            _name: &'static str,
                                            _variant: u32,
                                            variant_name: &'static str,
                                            value: &T) ->
        Result<Self::Ok> where T: Serialize
    {
        Ok(Value::simple_record(variant_name, vec![to_value(value)]).wrap())
    }

    fn serialize_seq(self, count: Option<usize>) -> Result<Self::SerializeSeq> {
        let vec = match count { Some(n) => Vec::with_capacity(n), None => Vec::new() };
        Ok(SerializeSequence{ vec, phantom: PhantomData })
    }

    fn serialize_tuple(self, count: usize) -> Result<Self::SerializeTuple> {
        Ok(SerializeRecord{ name: "tuple", vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_tuple_struct(self, name: &'static str, count: usize) ->
        Result<Self::SerializeTupleStruct>
    {
        Ok(SerializeRecord{ name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_tuple_variant(self,
                               _name: &'static str,
                               _variant: u32,
                               variant_name: &'static str,
                               count: usize) ->
        Result<Self::SerializeTupleVariant>
    {
        Ok(SerializeRecord{ name: variant_name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_map(self, _count: Option<usize>) -> Result<Self::SerializeMap> {
        Ok(SerializeDictionary{ next_key: None, items: Map::new(), phantom: PhantomData })
    }

    fn serialize_struct(self, name: &'static str, count: usize) -> Result<Self::SerializeStruct> {
        Ok(SerializeRecord{ name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }

    fn serialize_struct_variant(self,
                                _name: &'static str,
                                _variant: u32,
                                variant_name: &'static str,
                                count: usize) ->
        Result<Self::SerializeStructVariant>
    {
        Ok(SerializeRecord{ name: variant_name, vec: Vec::with_capacity(count), phantom: PhantomData })
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeMap for SerializeDictionary<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_key<T: ?Sized>(&mut self, key: &T) -> Result<()> where T: Serialize {
        self.next_key = Some(to_value(key));
        Ok(())
    }

    fn serialize_value<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        let key = self.next_key.take().unwrap();
        self.items.insert(key, to_value(value));
        Ok(())
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(Value::from(self.items).wrap())
    }
}

impl<N: NestedValue<D>, D: Domain> SerializeRecord<N, D> {
    fn push(&mut self, value: N) -> Result<()> {
        self.vec.push(value);
        Ok(())
    }

    fn finish(self) -> Result<N> {
        Ok(Value::simple_record(self.name, self.vec).wrap())
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeStruct for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        self.push(to_value(value))
    }

    fn end(self) -> Result<Self::Ok> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeStructVariant for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        self.push(to_value(value))
    }

    fn end(self) -> Result<Self::Ok> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeTuple for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.push(to_value(value))
    }

    fn end(self) -> Result<Self::Ok> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeTupleStruct for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.push(to_value(value))
    }

    fn end(self) -> Result<Self::Ok> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeTupleVariant for SerializeRecord<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.push(to_value(value))
    }

    fn end(self) -> Result<Self::Ok> {
        self.finish()
    }
}

impl<N: NestedValue<D>, D: Domain> serde::ser::SerializeSeq for SerializeSequence<N, D> {
    type Ok = N;
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.vec.push(to_value(value));
        Ok(())
    }

    fn end(self) -> Result<Self::Ok> {
        Ok(Value::from(self.vec).wrap())
    }
}

pub fn to_value<T, N, D: Domain>(value: T) -> N where T: Serialize, N: NestedValue<D> {
    value.serialize(Serializer{ phantom: PhantomData }).unwrap()
}
